# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from abc import ABCMeta
from abc import abstractproperty
from collections import Mapping
from collections import MutableMapping
from collections import defaultdict
from collections import namedtuple
from datetime import date
from datetime import datetime
from functools import partial
from uuid import UUID

from six import iteritems
from six import iterkeys
from six import text_type
from six import with_metaclass

from m3_gar_client.constants import GAR_LEVELS


FieldDescriptor = namedtuple('FieldDescriptor', [
    'data_type',
    'required',
    'description',
])


class ReadOnlyAttribute(object):

    def __init__(self, name, doc=None):
        self._name = name
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._name] if obj else self

    def __set__(self, obj, value):
        raise AttributeError('{} is read only attribute'.format(self._name))

    def __delete__(self, obj):
        raise AttributeError('{} is read only attribute'.format(self._name))


class ObjectMeta(ABCMeta):

    def __new__(mcs, name, bases, namespace, **kwargs):  # @NoSelf
        cls = ABCMeta.__new__(mcs, name, bases, namespace)

        fields = namespace['fields']
        if isinstance(fields, dict):
            for field_name, field_descriptor in iteritems(fields):
                setattr(cls, field_name, ReadOnlyAttribute(
                    name=field_name,
                    doc=field_descriptor.description,
                ))

        return cls


class ObjectBase(with_metaclass(ObjectMeta, object)):
    """
    Базовый класс для объектов ГАР.
    Обеспечивает неизменяемость данных (объекты только для чтения).
    """

    @abstractproperty
    def fields(self):
        """Описание полей объекта."""

    def __hash__(self):
        return hash(self.id)  # pylint: disable=no-member

    def __init__(self, **kwargs):
        for field_name, field_descriptor in iteritems(self.fields):
            if field_descriptor.required:
                if field_name in kwargs:
                    field_value = kwargs[field_name]
                else:
                    raise TypeError('argument {} is required'.format(field_name))
            else:
                field_value = kwargs.get(field_name, None)

            if field_value and field_descriptor.data_type:
                try:
                    field_value = field_descriptor.data_type(field_value)
                except ValueError:
                    raise ValueError('{} = {}'.format(field_name, field_value))

            self.__dict__[field_name] = field_value


# dict в списке базовых классов нужен для правильной обработки данных в M3.
class ObjectDictAdapter(Mapping, dict):
    """
    Адаптер для объектов ГАР, преобразующий их к словарям.
    """

    def __init__(self, obj):  # pylint: disable=super-init-not-called
        """Инициализация экземпляра класса.

        :type obj: m3_gar_client.data.ObjectBase
        """
        assert isinstance(obj, ObjectBase), type(obj)

        self._obj = obj

    def __iter__(self):
        return iter(self._obj.fields)

    def __getitem__(self, key):
        return getattr(self._obj, key)

    def __len__(self):
        return len(self._obj.fields)

    # pylint: disable=unused-argument
    @property
    def __readonly_exception(self):
        return TypeError("'{}' object is readonly".format(self.__class__.__name__))

    def __setitem__(self, key, value):
        raise self.__readonly_exception

    def __delitem__(self, key):
        raise self.__readonly_exception

    def pop(self, *args):
        raise self.__readonly_exception

    def popitem(self):
        raise self.__readonly_exception

    def clear(self):
        raise self.__readonly_exception

    def update(self, *args, **kwargs):
        raise self.__readonly_exception

    def setdefault(self, key, default=None):
        raise self.__readonly_exception

    # pylint: enable=unused-argument


def _unicode(value):
    return value if isinstance(value, text_type) else text_type(value)


def _unicode_or_empty(value):
    return _unicode(value) if value else ''


def _int(value):
    try:
        return int(value if isinstance(value, int) else int(value))
    except (TypeError, ValueError):
        raise ValueError(value)


def _int_or_none(value):
    return _int(value) if value else None


def _uuid(value):
    try:
        return text_type(value if isinstance(value, UUID) else UUID(value))
    except (AttributeError, TypeError, ValueError):
        raise ValueError(value)


def _uuid_or_none(value):
    return _uuid(value) if value else None


def _uuid_or_int(value):
    try:
        result = _int(value)
    except ValueError:
        result = _uuid(value)

    return result


def _uuid_or_int_or_none(value):
    return _uuid_or_int(value) if value else None


def _date(value):
    if isinstance(value, date):
        result = value
    else:
        try:
            result = datetime.strptime(value, '%Y-%m-%d').date()
        except (ValueError, TypeError):
            raise ValueError(value)

    return result


def _choices(valid_values, value_type, raw_value):
    value = value_type(raw_value)

    if value not in valid_values:
        raise ValueError(raw_value)

    return value


class AddressObject(ObjectBase):
    """
    Адресный объект.
    """

    #: Поля объекта и признак обязательности значения.
    fields = {
        'id': FieldDescriptor(
            data_type=_int,
            required=True,
            description='Уникальный идентификатор записи (ключевое поле).',
        ),
        'previous_id': FieldDescriptor(
            data_type=_int,
            required=False,
            description='Идентификатор записи связывания с предыдушей исторической записью.',
        ),
        'next_id': FieldDescriptor(
            data_type=_int,
            required=False,
            description='Идентификатор записи  связывания с последующей исторической записью.',
        ),
        'obj_id': FieldDescriptor(
            data_type=_int,
            required=True,
            description='Глобальный уникальный идентификатор адресного объекта.',
        ),
        'guid': FieldDescriptor(
            data_type=_uuid,
            required=True,
            description='UUID адресного объекта.',
        ),
        'level': FieldDescriptor(
            data_type=partial(_choices, GAR_LEVELS, _int),
            required=True,
            description='Уровень адресного объекта.',
        ),
        'ifns_fl_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код ИФНС для физических лиц.',
        ),
        'ifns_fl_territorial_district_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код территориального участка ИФНС ФЛ.',
        ),
        'ifns_ul_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код ИФНС для юридических лиц.',
        ),
        'ifns_ul_territorial_district_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код территориального участка ИФНС ЮЛ.',
        ),
        'okato': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='ОКАТО.',
        ),
        'oktmo': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='ОКТМО.',
        ),
        'kladr_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код адресного объекта одной строкой с признаком актуальности из КЛАДР 4.0.',
        ),
        'kladr_plain_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код адресного объекта из КЛАДР 4.0 одной строкой без признака актуальности (последних двух цифр).',
        ),
        'short_name': FieldDescriptor(
            data_type=_unicode,
            required=True,
            description='Краткое наименование типа объекта',
        ),
        'official_name': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Официальное наименование',
        ),
        'formal_name': FieldDescriptor(
            data_type=_unicode,
            required=True,
            description='Формализованное наименование',
        ),
        'postal_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Почтовый индекс',
        ),
        'date_of_creation': FieldDescriptor(
            data_type=_date,
            required=True,
            description='Начало действия записи',
        ),
        'date_of_update': FieldDescriptor(
            data_type=_date,
            required=True,
            description='Дата  внесения (обновления) записи',
        ),
        'date_of_expiration': FieldDescriptor(
            data_type=_date,
            required=True,
            description='Окончание действия записи',
        ),
        'full_name': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Полный адрес',
        ),
        'region_code': FieldDescriptor(
            data_type=_int,
            required=True,
            description='Код региона',
        ),
        'adm_parent_obj_id': FieldDescriptor(
            data_type=_int,
            required=False,
            description='Родительский obj_id в административной иерархии'
        ),
        'adm_parent_guid': FieldDescriptor(
            data_type=_uuid,
            required=False,
            description='Родительский guid в административной иерархии'
        ),
        'mun_parent_obj_id': FieldDescriptor(
            data_type=_int,
            required=False,
            description='Родительский obj_id в муниципальной иерархии'
        ),
        'mun_parent_guid': FieldDescriptor(
            data_type=_uuid,
            required=False,
            description='Родительский guid в муниципальной иерархии'
        ),
    }


class House(ObjectBase):
    """
    Объект здания (дома).
    """

    fields = {
        'id': FieldDescriptor(
            data_type=_int,
            required=True,
            description='Уникальный идентификатор записи дома',
        ),
        'obj_id': FieldDescriptor(
            data_type=_int,
            required=True,
            description='Глобальный уникальный идентификатор дома',
        ),
        'guid': FieldDescriptor(
            data_type=_uuid,
            required=True,
            description='UUID адресного объекта.',
        ),
        'ifns_fl_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код ИФНС для физических лиц.',
        ),
        'ifns_fl_territorial_district_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код территориального участка ИФНС ФЛ.',
        ),
        'ifns_ul_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код ИФНС для юридических лиц.',
        ),
        'ifns_ul_territorial_district_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Код территориального участка ИФНС ЮЛ.',
        ),
        'okato': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='ОКАТО.',
        ),
        'oktmo': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='ОКТМО.',
        ),
        'postal_code': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Почтовый индекс',
        ),
        'house_number': FieldDescriptor(
            data_type=_unicode,
            required=True,
            description='Номер дома',
        ),
        'building_number': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Номер корпуса',
        ),
        'structure_number': FieldDescriptor(
            data_type=_unicode_or_empty,
            required=False,
            description='Номер строения',
        ),
        'date_of_creation': FieldDescriptor(
            data_type=_date,
            required=True,
            description='Начало действия записи',
        ),
        'date_of_update': FieldDescriptor(
            data_type=_date,
            required=True,
            description='Дата  внесения (обновления) записи',
        ),
        'date_of_end': FieldDescriptor(
            data_type=_date,
            required=True,
            description='Окончание действия записи',
        ),
        'region_code': FieldDescriptor(
            data_type=_int,
            required=True,
            description='Код региона',
        ),
        'adm_parent_obj_id': FieldDescriptor(
            data_type=_int,
            required=False,
            description='Родительский obj_id в административной иерархии'
        ),
        'adm_parent_guid': FieldDescriptor(
            data_type=_uuid,
            required=False,
            description='Родительский guid в административной иерархии'
        ),
        'mun_parent_obj_id': FieldDescriptor(
            data_type=_int,
            required=False,
            description='Родительский obj_id в муниципальной иерархии'
        ),
        'mun_parent_guid': FieldDescriptor(
            data_type=_uuid,
            required=False,
            description='Родительский guid в муниципальной иерархии'
        ),
    }


class ObjectMapper(with_metaclass(ABCMeta, MutableMapping)):
    """
    Обертка над словарями, преобразующая ключи.
    """

    @abstractproperty
    def fields_map(self):
        """Список соответствия ключей.

        :rtype: dict
        """

    def __init__(self, data):  # pylint: disable=super-init-not-called
        assert isinstance(data, dict), type(data)

        self._data = data

    def __len__(self):
        i = 0

        for i, _ in enumerate(iterkeys(self), 1):
            pass

        return i

    def __iter__(self):
        return (field for field in self.fields_map)

    def __getitem__(self, key):
        field_source = self.fields_map[key]

        if callable(field_source):
            # могут быть поля, значение которых не получается вытащить из одного атрибута.
            # для них в fields_map указываем функцию, принимающую весь data единственным аргументом
            value = field_source(self._data)
        else:
            value = self._data.get(field_source)

        return value

    def __setitem__(self, key, value):
        self._data[self.fields_map[key]] = value

    def __delitem__(self, key):
        del self._data[self.fields_map[key]]

    def __bool__(self):
        return bool(self._data)

    __nonzero__ = __bool__



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from importlib import import_module
import abc

from django.conf import settings
from six import with_metaclass

from m3_gar_client.utils import cached_property


class Config(with_metaclass(abc.ABCMeta, object)):

    """Базовый класс для конфигурации пакета."""

    @cached_property
    def backend(self):
        """Бэкенд для доступа к данным ГАР.

        :rtype: :class:`m3_gar_client.backends.base.BackendBase`
        """
        backend_class = import_module(settings.GAR['BACKEND']).Backend
        return backend_class()


#: Конфигурация приложения ``m3_gar_client``.
#:
#: Заполняется экземпляром класса :class:`m3_gar_client.Config`, либо его потомком,
#: при инициализации проекта *до* инициализации приложения ``m3-gar-client``.
config = None


if settings.DEBUG:
    import os
    os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import json
import warnings

from m3_ext.ui.base import BaseExtComponent
from m3_ext.ui.containers.containers import ExtContainer
from m3_ext.ui.fields.simple import ExtComboBox
from m3_ext.ui.fields.simple import ExtHiddenField
from m3_ext.ui.fields.simple import ExtStringField
from m3_ext.ui.fields.simple import ExtTextArea
from m3_ext.ui.misc.store import ExtJsonStore

from m3_gar_client.constants import UI_LEVEL_FLAT
from m3_gar_client.constants import UI_LEVEL_HOUSE
from m3_gar_client.constants import UI_LEVEL_PLACE
from m3_gar_client.constants import UI_LEVEL_STREET
from m3_gar_client.constants import UI_LEVELS
from m3_gar_client.data import ObjectDictAdapter
from m3_gar_client.data import ObjectMapper
from m3_gar_client.utils import cached_property
from m3_gar_client.utils import get_address_object
from m3_gar_client.utils import get_house
import m3_gar_client


class UIAddressObjectMapper(ObjectMapper):
    """Обертка над адресным объектом для передачи данных в UI.

    Преобразует наименования полей адресного объекта в наименования полей в
    компонентах ExtJS.
    """

    fields_map = {
        'objectId': 'guid',
        'level': 'level',
        'shortName': 'short_name',
        'formalName': 'formal_name',
        'postalCode': 'postal_code',
        'fullName': 'full_name',
    }


class UIHouseMapper(ObjectMapper):
    """Обертка над домом для передачи данных в UI.

    Преобразует наименования полей дома в наименования полей в компонентах
    ExtJS.
    """

    fields_map = {
        'objectId': 'guid',
        'houseNumber': 'house_number',
        'buildingNumber': 'building_number',
        'structureNumber': 'structure_number',
        'postalCode': 'postal_code',
    }


class AddressFields(BaseExtComponent):
    """Контейнер для полей редактирования составных элементов адреса.

    В зависимости от указанного уровня точности создает необходимые поля.

    Также реализует настройку поведения полей в зависимости от их заполнения.
    Например, если не заполнено поле "Населенный пункт", то поля "Улица",
    "Дом", "Корпус", "Строение", "Квартира" становятся недоступными для
    редактирования.

    Предназначен для использования в представлении адресной панели.
    """

    def __init__(self, *args, **kwargs):
        assert m3_gar_client.config.backend is not None
        self.backend = None

        # Флаг, определяющий обязательность заполнения адреса.
        self.allow_blank = True

        # Флаг, определяющий возможность редактирования адреса.
        self.read_only = False

        # Уровень точности адреса.
        self.level = UI_LEVEL_FLAT

        # Имена полей.
        self._names_of_fields = {
            'place_name': 'place_name',
            'place_id': 'place_id',
            'zip_code': 'zip_code',
            'street_name': 'street_name',
            'street_id': 'street_id',
            'house_number': 'house_number',
            'building_number': 'building_number',
            'structure_number': 'structure_number',
            'house_id': 'house_id',
            'flat_number': 'flat_number',
            'full_address': 'full_address',
        }

        # Подписи полей ввода.
        self._labels_text = {
            'place_name': 'Населенный пункт',
            'zip_code': 'Индекс',
            'street_name': 'Улица',
            'house_number': 'Дом',
            'building_number': 'Корпус',
            'structure_number': 'Строение',
            'flat_number': 'Квартира',
            'full_address': 'Полный адрес',
        }

        # Флаг, определяющий возможность ввода адресов, отсутствующих в ГАР.
        self.gar_only = True

        # Флаг, определяющий отображение поля с полным адресом.
        self.with_full_address = None

        # Timeout запросов к серверу ГАР.
        self.timeout = None

        super(AddressFields, self).__init__(*args, **kwargs)

        self.init_component(*args, **kwargs)

        self.backend = m3_gar_client.config.backend

        assert self.level in UI_LEVELS, self.level

        if self.with_full_address is None:
            self.with_full_address = self.level != UI_LEVEL_PLACE

    @property
    def names_of_fields(self):
        return self._names_of_fields

    @names_of_fields.setter
    def names_of_fields(self, value):
        self._names_of_fields.update(value)

    @property
    def labels_text(self):
        return self._labels_text

    @labels_text.setter
    def labels_text(self, value):
        self._labels_text.update(value)

    @cached_property
    def field__place_name(self):
        """Поле для ввода названия населенного пункта.

        :rtype: m3_ext.ui.fields.simple.ExtComboBox
        """
        result = ExtComboBox(
            name=self._names_of_fields['place_name'],
            label=self._labels_text['place_name'],
            display_field='fullName',
            value_field='fullName',
            query_param='filter',
            hide_trigger=True,
            force_selection=self.gar_only,
            min_chars=2,
            empty_text='Название субъекта/города/населенного пункта',
            read_only=self.read_only,
            allow_blank=self.allow_blank,
            fields=[
                'objectId',
                'level',
                'fullName',
                'postalCode',
            ],
            store=ExtJsonStore(
                url=self.backend.place_search_url,
                id_property='objectId',
                root='rows',
                total_property='total',
            ),
        )

        self.backend.configure_place_field(result)

        return result

    @cached_property
    def field__place_id(self):
        """Поле для хранения ID населенного пункта.

        :rtype: m3_ext.ui.fields.simple.ExtHiddenField
        """
        return ExtHiddenField(
            type=ExtHiddenField.STRING,
            name=self._names_of_fields['place_id'],
        )

    @cached_property
    def field__zip_code(self):
        """Поле для отображения/ввода почтового индекса.

        Если параметр ``gar_only`` равен ``True``, то редактирование значения
        поля будет недоступно.

        :rtype: m3_ext.ui.fields.simple.ExtStringField
        """
        return ExtStringField(
            name=self._names_of_fields['zip_code'],
            label=self._labels_text['zip_code'],
            read_only=self.gar_only or self.read_only,
            width=50,
        )

    @cached_property
    def field__street_name(self):
        """Поле для ввода названия улицы.

        :rtype: m3_ext.ui.fields.simple.ExtComboBox
        """
        result = ExtComboBox(
            name=self._names_of_fields['street_name'],
            label=self._labels_text['street_name'],
            display_field='name',
            value_field='name',
            query_param='filter',
            hide_trigger=True,
            force_selection=self.gar_only,
            min_chars=2,
            empty_text='Название улицы/микрорайона',
            read_only=self.read_only,
            fields=[
                'objectId',
                'level',
                'shortName',
                'postalCode',
                'formalName',
                'name',  # значение поля формируется как shortName + formalName
            ],
            store=ExtJsonStore(
                url=self.backend.street_search_url,
                id_property='objectId',
                root='rows',
                total_property='total',
            ),
        )

        self.backend.configure_street_field(result)

        return result

    @cached_property
    def field__street_id(self):
        """Поле для хранения ID улицы.

        :rtype: m3_ext.ui.fields.simple.ExtHiddenField
        """
        return ExtHiddenField(
            type=ExtHiddenField.STRING,
            name=self._names_of_fields['street_id'],
        )

    @cached_property
    def field__house_number(self):
        """Поле для ввода номера дома.

        :rtype: m3_ext.ui.fields.simple.ExtComboBox
        """
        result = ExtComboBox(
            name=self._names_of_fields['house_number'],
            label=self._labels_text['house_number'],
            display_field='houseNumber',
            value_field='houseNumber',
            query_param='filter',
            hide_trigger=True,
            force_selection=self.gar_only,
            min_chars=1,
            width=40,
            list_width=150,
            read_only=self.read_only,
            fields=[
                'objectId',
                'houseNumber',
                'buildingNumber',
                'structureNumber',
                'postalCode',
            ],
            store=ExtJsonStore(
                url=self.backend.house_search_url,
                id_property='objectId',
                root='rows',
                total_property='total',
            ),
        )

        self.backend.configure_house_field(result)

        return result

    @cached_property
    def field__building_number(self):
        """Поле для ввода номера корпуса.

        :rtype: m3_ext.ui.fields.simple.ExtComboBox
        """
        result = ExtStringField(
            name=self._names_of_fields['building_number'],
            label=self._labels_text['building_number'],
            read_only=self.gar_only or self.read_only,
            width=40,
        )

        return result

    @cached_property
    def field__structure_number(self):
        """Поле для ввода номера строения.

        :rtype: m3_ext.ui.fields.simple.ExtComboBox
        """
        result = ExtStringField(
            name=self._names_of_fields['structure_number'],
            label=self._labels_text['structure_number'],
            read_only=self.gar_only or self.read_only,
            width=40,
        )

        return result

    @cached_property
    def field__house_id(self):
        """Поле для хранения ID дома.

        :rtype: m3_ext.ui.fields.simple.ExtHiddenField
        """
        return ExtHiddenField(
            type=ExtHiddenField.STRING,
            name=self._names_of_fields['house_id'],
        )

    @cached_property
    def field__flat_number(self):
        """Поле для ввода номера квартиры.

        :rtype: m3_ext.ui.fields.simple.ExtStringField
        """
        return ExtStringField(
            name=self._names_of_fields['flat_number'],
            label=self._labels_text['flat_number'],
            width=40,
            read_only=self.read_only
        )

    @cached_property
    def field__full_address(self):
        """Поле для отображения/ввода полного адреса.

        Если параметр ``gar_only`` равен ``True``, то редактирование значения
        поля будет недоступно.

        :rtype: m3_ext.ui.fields.simple.ExtTextArea
        """
        return ExtTextArea(
            name=self._names_of_fields['full_address'],
            label=self._labels_text['full_address'],
            height=36,
            read_only=self.gar_only or self.read_only,
        )

    @property
    def items(self):
        """Возвращает все поля.

        Т.к. это не контейнер, в коде JavaScript этого параметра не будет, но
        он используется для биндинга формы с объектами.
        """
        result = [
            self.field__place_id,
            self.field__place_name,
            self.field__zip_code,
        ]

        if self.level in (UI_LEVEL_STREET, UI_LEVEL_HOUSE, UI_LEVEL_FLAT):
            result.extend((
                self.field__street_id,
                self.field__street_name,
            ))

        if self.level in (UI_LEVEL_HOUSE, UI_LEVEL_FLAT):
            result.extend((
                self.field__house_id,
                self.field__house_number,
                self.field__building_number,
                self.field__structure_number,
            ))

        if self.level == UI_LEVEL_FLAT:
            result.append(
                self.field__flat_number
            )

        if self.with_full_address:
            result.append(
                self.field__full_address
            )

        return result

    @cached_property
    def place(self):
        """Населенный пункт.

        :rtype: m3_gar_client.data.AddressObject
        """
        if self.field__place_id.value:
            return get_address_object(self.field__place_id.value, self.timeout)

    @cached_property
    def street(self):
        """Улица.

        :rtype: m3_gar_client.data.AddressObject
        """
        assert self.level in (UI_LEVEL_STREET, UI_LEVEL_HOUSE, UI_LEVEL_FLAT)

        if self.field__street_id.value:
            return get_address_object(self.field__street_id.value, self.timeout)

    @cached_property
    def house(self):
        """Дом.

        :rtype: m3_gar_client.data.House
        """
        assert self.level in (UI_LEVEL_HOUSE, UI_LEVEL_FLAT)

        if self.field__house_id.value and (self.field__street_id.value or self.field__place_id.value):
            return get_house(
                house_id=self.field__house_id.value,
                parent_id=self.field__street_id.value or self.field__place_id.value,
                timeout=self.timeout,
            )

    def find_by_name(self, name):
        """Поиск экземпляра поля по имени.

        Метод Ext-контейнеров, позволяющий рекурсивно искать вложенные элементы
        """
        for item in self.items:
            if hasattr(item, 'name') and name == getattr(item, 'name'):
                return item

    def render_base_config(self):
        super(AddressFields, self).render_base_config()

        # Поля ввода элементов адреса.
        put = self._put_config_value

        put('level', self.level)
        put('garOnly', self.gar_only)

        put('placeNameField', self.field__place_name.render)
        put('placeIDField', self.field__place_id.render)

        if self.place:
            place_dict = ObjectDictAdapter(self.place)
            put('place', dict(UIAddressObjectMapper(place_dict)))

        put('zipCodeField', self.field__zip_code.render)

        if self.level in (UI_LEVEL_STREET, UI_LEVEL_HOUSE, UI_LEVEL_FLAT):
            put('streetNameField', self.field__street_name.render)
            put('streetIDField', self.field__street_id.render)

            if self.street:
                street_dict = ObjectDictAdapter(self.street)
                put('street', dict(UIAddressObjectMapper(street_dict)))

        if self.level in (UI_LEVEL_HOUSE, UI_LEVEL_FLAT):
            put('houseNumberField', self.field__house_number.render)
            put('buildingNumberField', self.field__building_number.render)
            put('structureNumberField', self.field__structure_number.render)
            put('houseIDField', self.field__house_id.render)

            if self.house:
                house_dict = ObjectDictAdapter(self.house)
                put('house', dict(UIHouseMapper(house_dict)))

        if self.level == UI_LEVEL_FLAT:
            put('flatNumberField', self.field__flat_number.render)

        if self.with_full_address:
            put('withFullAddress', self.with_full_address)
            put('fullAddressField', self.field__full_address.render)

    def render(self):
        self.pre_render()
        self.render_base_config()

        return 'new Ext.m3.gar.AddressFields({%s})' % self._get_config_str()


class AddressViewBase(ExtContainer):
    """Базовый класс для представлений панели ввода адреса."""

    def _get_labels_width(self):  # pylint: disable=no-self-use
        return {}

    def __init__(self, *args, **kwargs):
        self._labels_width = self._get_labels_width()
        self.fields = None

        super(AddressViewBase, self).__init__(*args, **kwargs)

        assert isinstance(self.fields, AddressFields), type(self.fields)

    @property
    def labels_width(self):
        return self._labels_width

    @labels_width.setter
    def labels_width(self, value):
        self._labels_width.update(value)

    def render_base_config(self):
        super(AddressViewBase, self).render_base_config()

        put = self._put_config_value
        put('labelsWidth', self._labels_width)

    def render(self):
        self.pre_render()
        self.render_base_config()

        return 'new %s({%s})' % (self._ext_name, self._get_config_str())


class RowsAddressView(AddressViewBase):
    """Представление адресной панели с размещением элементов на трех строках.

    Каждая из строк содержит следующие поля:

        1. Населенный пункт, Индекс
        2. Улица, Дом, Корпус, Строение, Квартира
        3. Полный адрес

    Поля "Населенный пункт", "Улица" и "Полный адрес" выровнены слева.
    """

    def _get_labels_width(self):
        result = super(RowsAddressView, self)._get_labels_width()

        result.update({
            'place': 110,
            'zipCode': 44,
            'street': 110,
            'house': 28,
            'building': 44,
            'structure': 58,
            'flat': 55,
            'fullAddress': 110,
        })

        return result

    def __init__(self, *args, **kwargs):
        super(RowsAddressView, self).__init__(*args, **kwargs)

        self._ext_name = 'Ext.m3.gar.RowsAddressView'

    def init_component(self, *args, **kwargs):
        super(RowsAddressView, self).init_component(*args, **kwargs)

        # Контейнер полей добавляется в items для того, чтобы была возможность
        # биндинга объекта с полями формы. В JavaScript этот объект будет
        # удален из items.
        self.items.append(self.fields)


class CompactAddressView(RowsAddressView):

    def _get_labels_width(self):
        result = super(CompactAddressView, self)._get_labels_width()

        result.update({
            'place': 110,
            'street': 38,
            'fullAddress': 87,
        })

        return result


class AdvancedRowsAddressView(RowsAddressView):
    """
    Перекрытая реализация адресного компонента ГАР
    """

    def __init__(self, *args, **kwargs):
        if 'fields' not in kwargs:
            fields = AddressFields()

            if 'gar_only' in kwargs:
                fields.gar_only = kwargs['gar_only']
                del kwargs['gar_only']

            kwargs['fields'] = fields

        super(AdvancedRowsAddressView, self).__init__(*args, **kwargs)
        self._ext_name = 'Ext.m3.gar.NewAddressView'

    def make_read_only(self, access_off=True, exclude_list=(), *args, **kwargs):
        result = super().make_read_only(access_off, exclude_list, *args, **kwargs)

        self.fields.read_only = access_off

        return result

    def _get_labels_width(self):
        result = super()._get_labels_width()

        result.update({
            'house': 110,
        })

        return result

    def get_use_corps(self):
        warnings.warn(
            "Использование атрибута `use_corps` более не требуется и рекомендуется к удалению",
            DeprecationWarning,
        )

        return True

    def set_use_corps(self, value):
        warnings.warn(
            "Использование атрибута `use_corps` более не требуется и рекомендуется к удалению",
            DeprecationWarning,
        )

    # legacy поле (сохранено для совместимости)
    use_corps = property(get_use_corps, set_use_corps)

    def get_addr_field_name(self):
        return self.fields.names_of_fields['full_address']

    def set_addr_field_name(self, value):
        self.fields.names_of_fields['full_address'] = value

    addr_field_name = property(get_addr_field_name, set_addr_field_name)

    def get_flat_field_name(self):
        return self.fields.names_of_fields['flat_number']

    def set_flat_field_name(self, value):
        self.fields.names_of_fields['flat_number'] = value

    flat_field_name = property(get_flat_field_name, set_flat_field_name)

    def get_street_field_name(self):
        return self.fields.names_of_fields['street_id']

    def set_street_field_name(self, value):
        self.fields.names_of_fields['street_id'] = value

    street_field_name = property(get_street_field_name, set_street_field_name)

    def get_place_field_name(self):
        return self.fields.names_of_fields['place_id']

    def set_place_field_name(self, value):
        self.fields.names_of_fields['place_id'] = value

    place_field_name = property(get_place_field_name, set_place_field_name)

    def get_house_field_name(self):
        return self.fields.names_of_fields['house_number']

    def set_house_field_name(self, value):
        self.fields.names_of_fields['house_number'] = value

    house_field_name = property(get_house_field_name, set_house_field_name)

    def get_zipcode_field_name(self):
        return self.fields.names_of_fields['zip_code']

    def set_zipcode_field_name(self, value):
        self.fields.names_of_fields['zip_code'] = value

    zipcode_field_name = property(get_zipcode_field_name, set_zipcode_field_name)

    def get_corps_field_name(self):
        return self.fields.names_of_fields['building_number']

    def set_corps_field_name(self, value):
        self.fields.names_of_fields['building_number'] = value

    corps_field_name = property(get_corps_field_name, set_corps_field_name)

    def get_structure_field_name(self):
        return self.fields.names_of_fields['structure_number']

    def set_structure_field_name(self, value):
        self.fields.names_of_fields['structure_number'] = value

    structure_field_name = property(get_structure_field_name, set_structure_field_name)

    @classmethod
    def addr_object(cls, code):
        """
        Информация об адресном объекте для ручного биндинга значений в поля
        """
        obj = get_address_object(code)

        data = {
            'address': '',
            'ao_guid': '',
            'ao_level': '',
            'formal_name': '',
            'name': '',
            'place_address': '',
            'postal_code': '',
            'shortname': '',
        }

        if obj:
            data.update({
                'ao_guid': str(obj.guid),
                'ao_level': obj.level,
                'formal_name': obj.formal_name,
                'name': obj.formal_name,
                'postal_code': obj.postal_code,
                'shortname': obj.short_name,
            })

        return json.dumps(data)

    def get_zipcode(self):
        """
        Получаем значение почтового индекса. Если он явно не указан, то берем
        первые 6 символов текстового представления адреса и пытаемся проверить,
        являются ли они числами. В случае успеха, возвращаем их.
        """
        result = ''

        if self.field__zip_code and self.field__zip_code.value:
            result = self.field__zip_code.value

        if self.field__full_address:
            addr = self.field__full_address.value

            if len(addr) > 6:
                try:
                    result = str(int(addr[:6]))
                except ValueError:
                    pass

        return result



# coding: utf-8

# -----------------------------------------------------------------------------
# Уровень точности адреса: населенный пункт.
UI_LEVEL_PLACE = 1

# Уровень точности адреса: улица.
UI_LEVEL_STREET = 2

# Уровень точности адреса: дом.
UI_LEVEL_HOUSE = 3

# Уровень точности адреса: квартира.
UI_LEVEL_FLAT = 4

# Уровни точности адреса
UI_LEVELS = (
    UI_LEVEL_PLACE,
    UI_LEVEL_STREET,
    UI_LEVEL_HOUSE,
    UI_LEVEL_FLAT,
)

# -----------------------------------------------------------------------------
# Уровни адресных объектов ГАР.

# Субъект РФ
GAR_LEVEL_RF_REGION = 1

# Административный район
GAR_LEVEL_ADM_DISTRICT = 2

# Муниципальный район
GAR_LEVEL_MUN_DISTRICT = 3

# Сельское/городское поселение
GAR_LEVEL_SETTLEMENT = 4

# Город
GAR_LEVEL_CITY = 5

# Населенный пункт
GAR_LEVEL_LOCALITY = 6

# Элемент планировочной структуры
GAR_LEVEL_PLANNING_STRUCTURE = 7

# Элемент улично-дорожной сети
GAR_LEVEL_STREET = 8

# Земельный участок
GAR_LEVEL_STEAD = 9

# Здание (сооружение)
GAR_LEVEL_BUILDING = 10

# Помещение
GAR_LEVEL_ROOM = 11

# Помещения в пределах помещения
GAR_LEVEL_ROOM_IN_ROOM = 12

# Автономный округ (устаревшее)
GAR_LEVEL_AUTONOMOUS_DISTRICT = 13

# Внутригородская территория (устаревшее)
GAR_LEVEL_INTRACITY_TERRITORY = 14

# Дополнительные территории (устаревшее)
GAR_LEVEL_ADDITIONAL_TERRITORY = 15

# Объекты на дополнительных территориях (устаревшее)
GAR_LEVEL_ADDITIONAL_TERRITORY_OBJECT = 16

# Машино-место (устаревшее)
GAR_LEVEL_PARKING_PLACE = 17


# Уровни адресных объектов
GAR_LEVELS = (
    GAR_LEVEL_RF_REGION,
    GAR_LEVEL_ADM_DISTRICT,
    GAR_LEVEL_MUN_DISTRICT,
    GAR_LEVEL_SETTLEMENT,
    GAR_LEVEL_CITY,
    GAR_LEVEL_LOCALITY,
    GAR_LEVEL_PLANNING_STRUCTURE,
    GAR_LEVEL_STREET,
    GAR_LEVEL_STEAD,
    GAR_LEVEL_BUILDING,
    GAR_LEVEL_ROOM,
    GAR_LEVEL_ROOM_IN_ROOM,
    GAR_LEVEL_AUTONOMOUS_DISTRICT,
    GAR_LEVEL_INTRACITY_TERRITORY,
    GAR_LEVEL_ADDITIONAL_TERRITORY,
    GAR_LEVEL_ADDITIONAL_TERRITORY_OBJECT,
    GAR_LEVEL_PARKING_PLACE,
)

# Уровни адресных объектов, относящихся к населенным пунктам.
GAR_LEVELS_PLACE = (
    GAR_LEVEL_RF_REGION,
    GAR_LEVEL_ADM_DISTRICT,
    GAR_LEVEL_MUN_DISTRICT,
    GAR_LEVEL_SETTLEMENT,
    GAR_LEVEL_CITY,
    GAR_LEVEL_LOCALITY,
    GAR_LEVEL_PLANNING_STRUCTURE,
)

# Уровни адресных объектов, относящимся к улицам.
GAR_LEVELS_STREET = (
    GAR_LEVEL_STREET,
)


# -----------------------------------------------------------------------------
# Виды иерархического деления ГАР.

# Иерархия в муниципальном делении
GAR_HIERARCHY_MUN = 'mun'

# Иерархия в административном делении
GAR_HIERARCHY_ADM = 'adm'


# Виды иерархического деления
GAR_HIERARCHIES = (
    GAR_HIERARCHY_MUN,
    GAR_HIERARCHY_ADM,
)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from abc import ABCMeta
from abc import abstractmethod
import hashlib
import json

from django.utils.functional import SimpleLazyObject
from requests import Session
from six import with_metaclass
from six.moves import http_client

from m3_gar_client.utils import cached_property


class ServerBase(with_metaclass(ABCMeta, object)):
    """
    Базовый класс для серверов ГАР.
    """

    def __init__(self, **kwargs):
        self._base_url = kwargs['url']
        self._timeout = kwargs.get('timeout')

    @property
    def base_url(self):
        return self._base_url

    @property
    @abstractmethod
    def _session(self):
        """HTTP-сессия с сервером m3-rest-gar.

        :rtype: requests.sessions.Session
        """

    def get(self, path, params=None, timeout=None):
        """Возвращает ответ на HTTP-запрос к API сервера ГАР.

        :rtype: requests.models.Response
        """
        response = self._session.get(
            self.base_url.rstrip('/') + path,
            params=params or {},
            timeout=timeout or self._timeout,
        )
        return response


class CachingMixin(object):
    """Класс-примесь для кэширования ответов на запросы.

    Параметры:

        * ``cache`` --- объект кэша. Рекомендуется использовать
          ```django.core.cache.cache`.
        * ``cache_key_prefix`` --- префикс для ключей в кэше.
        * ``cache_timeout`` --- длительность кэширования (в секундах).
    """

    def __init__(self, **kwargs):
        super(CachingMixin, self).__init__(**kwargs)

        self._cache = kwargs['cache']
        self._cache_key_prefix = kwargs.get('cache_key_prefix', 'm3-gar')
        self._cache_timeout = kwargs.get('cache_timeout', 24 * 60 * 60)

    def get(self, path, params=None, timeout=None):
        hasher = hashlib.sha1()
        hasher.update(':'.join((
            self._cache_key_prefix,
            path,
            json.dumps(params, sort_keys=True),
        )).encode('utf-8'))
        cache_key = hasher.hexdigest()

        if cache_key in self._cache:
            response = self._cache.get(cache_key)
        else:
            response = super(CachingMixin, self).get(path, params, timeout)

            if response.status_code == http_client.OK:
                self._cache.set(cache_key, response,)

        return response


class SimpleServer(ServerBase):
    """Сервер ГАР без аутентификации.

    Параметры:

        * ``url`` --- URL API сервера ГАР.
        * ``timeout`` --- timeout запроса к серверу ГАР в секундах.
    """

    @cached_property
    def _session(self):
        result = Session()

        result.trust_env = True

        return result


class SimpleCachingServer(CachingMixin, SimpleServer):
    """
    Сервер ГАР с кешированием без аутентификации.

    Параметры:

        * ``url`` --- URL API сервера ГАР.
        * ``timeout`` --- timeout запроса к серверу ГАР в секундах.
        * ``cache`` --- объект кэша. Рекомендуется использовать
          ```django.core.cache.cache`.
        * ``cache_key_prefix`` --- префикс для ключей в кэше.
        * ``cache_timeout`` --- длительность кэширования (в секундах).
    """


class OAuth2Server(ServerBase):  # pragma: no cover
    """Сервер ГАР с аутентификацией OAuth2.

    Параметры:

        * ``url`` --- URL API сервера ГАР.
        * ``timeout`` --- timeout запроса к серверу ГАР в секундах.
        * ``token_url`` --- Token endpoint URL, must use HTTPS.
        * ``client_id``.
        * ``username`` --- Username used by LegacyApplicationClients..
        * ``password`` --- Password used by LegacyApplicationClients..
        * ``client_secret``.

    .. seealso::

       :meth:`requests_oauthlib.OAuth2Session.fetch_token`
    """

    def __init__(self, **kwargs):
        super(OAuth2Server, self).__init__(**kwargs)

        self.token_url = kwargs['token_url']
        self.client_id = kwargs['client_id']
        self.client_secret = kwargs['client_secret']
        self.username = kwargs['username']
        self.password = kwargs['password']

    @cached_property
    def _session(self):
        from oauthlib.oauth2 import LegacyApplicationClient
        from requests_oauthlib import OAuth2Session

        result = OAuth2Session(
            client=LegacyApplicationClient(self.client_id)
        )
        result.trust_env = True
        result.fetch_token(
            token_url=self.token_url,
            username=self.username,
            password=self.password,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

        return result


class OAuth2CachingServer(CachingMixin, OAuth2Server):
    """Сервер ГАР с кешированием и аутентификацией OAuth2.

    Параметры:

        * ``url`` --- URL API сервера ГАР.
        * ``timeout`` --- timeout запроса к серверу ГАР в секундах.
        * ``token_url`` --- Token endpoint URL, must use HTTPS.
        * ``client_id``.
        * ``username`` --- Username used by LegacyApplicationClients..
        * ``password`` --- Password used by LegacyApplicationClients..
        * ``client_secret``.
        * ``cache`` --- объект кэша. Рекомендуется использовать
          ```django.core.cache.cache`.
        * ``cache_key_prefix`` --- префикс для ключей в кэше.
        * ``cache_timeout`` --- длительность кэширования (в секундах).

    .. seealso::

       :meth:`requests_oauthlib.OAuth2Session.fetch_token`
    """


def get_server():
    """Возвращает сервер ГАР, созданный в соответствии с настройками m3-gar-client.

    Параметры подключения к серверу m3-rest-gar должны быть размещены в
    настройках Django (``django.conf.settings``) в параметре ``GAR``, который
    должен содержать словарь со следующими ключами:

        - ``URL`` --- URL API сервера ГАР.
        - ``TIMEOUT`` --- timeout запроса к серверу ГАР в секундах.
        - ``USE_CACHE`` --- определяет необходимость кеширования HTTP-запросов
          к серверу m3-rest-gar. Значение по умолчанию: ``False``
        - ``OAUTH2`` --- параметры OAuth2 (если не указан, то аутентификация
          не используется):

          - ``TOKEN_URL`` --- Token endpoint URL, must use HTTPS.
          - ``CLIENT_ID``
          - ``CLIENT_SECRET``
          - ``USERNAME``
          - ``PASSWORD``

    :rtype: m3_gar_client.backends.m3_rest_gar.ServerBase
    """
    from django.conf import settings

    if settings.GAR.get('USE_CACHE', False):
        from django.core.cache import cache

        if 'OAUTH2' in settings.GAR:  # pragma: no cover
            result = OAuth2CachingServer(
                url=settings.GAR['URL'],
                timeout=settings.GAR.get('TIMEOUT'),
                cache=cache,
                token_url=settings.GAR['OAUTH2']['TOKEN_URL'],
                client_id=settings.GAR['OAUTH2']['CLIENT_ID'],
                client_secret=settings.GAR['OAUTH2'].get('CLIENT_SECRET'),
                username=settings.GAR['OAUTH2'].get('USERNAME'),
                password=settings.GAR['OAUTH2'].get('PASSWORD'),
            )

        else:
            result = SimpleCachingServer(
                url=settings.GAR['URL'],
                timeout=settings.GAR.get('TIMEOUT'),
                cache=cache,
            )

    else:
        if 'OAUTH2' in settings.GAR:  # pragma: no cover
            result = OAuth2Server(
                url=settings.GAR['URL'],
                timeout=settings.GAR.get('TIMEOUT'),
                token_url=settings.GAR['OAUTH2']['TOKEN_URL'],
                client_id=settings.GAR['OAUTH2']['CLIENT_ID'],
                client_secret=settings.GAR['OAUTH2'].get('CLIENT_SECRET'),
                username=settings.GAR['OAUTH2'].get('USERNAME'),
                password=settings.GAR['OAUTH2'].get('PASSWORD'),
            )

        else:
            result = SimpleServer(
                url=settings.GAR['URL'],
                timeout=settings.GAR.get('TIMEOUT'),
            )

    return result


server = SimpleLazyObject(get_server)



# coding: utf-8
"""Бэкенд, проксирующий запросы через веб-приложение."""
from __future__ import absolute_import
from __future__ import unicode_literals

from uuid import UUID

from six import text_type

from m3_gar_client.backends.base import BackendBase
from m3_gar_client.backends.m3_rest_gar.proxy.utils import find_address_objects
from m3_gar_client.backends.m3_rest_gar.proxy.utils import find_house
from m3_gar_client.backends.m3_rest_gar.proxy.utils import get_address_object
from m3_gar_client.backends.m3_rest_gar.proxy.utils import get_house


class Backend(BackendBase):
    """
    Бэкенд для работы с сервером m3-rest-gar.
    """

    def __init__(self, *args, **kwargs):
        super(Backend, self).__init__(*args, **kwargs)

        self._pack = None

    @staticmethod
    def _register_parsers():
        """Регистрация парсеров для параметров контекста."""
        from m3.actions.context import DeclarativeActionContext

        params = (
            (
                'm3-gar:unicode-or-none',
                lambda s: text_type(s) if s else None
            ),
            (
                'm3-gar:int-list',
                lambda s: [int(x) for x in s.split(',')]
            ),
            (
                'm3-gar:guid-or-none',
                lambda x: UUID(x) if x else None
            ),
        )

        for name, parser in params:
            DeclarativeActionContext.register_parser(name, parser)

    def register_packs(self):
        """Регистрирует наборы действий в M3."""
        from m3_gar_client import config
        from m3_gar_client.backends.m3_rest_gar.proxy.actions import Pack

        self._register_parsers()

        self._pack = Pack()

        config.controller.extend_packs((
            self._pack,
        ))

    def place_search_url(self):
        """URL для поиска населенных пунктов.

        :rtype: str
        """
        return self._pack.place_search_action.get_absolute_url()

    def street_search_url(self):
        """URL для поиска улиц.

        :rtype: str
        """
        return self._pack.street_search_action.get_absolute_url()

    def house_search_url(self):
        """URL для запроса списка домов.

        :rtype: str
        """
        return self._pack.house_search_action.get_absolute_url()

    def find_address_objects(
        self,
        filter_string,
        levels=None,
        typenames=None,
        parent_id=None,
        timeout=None,
    ):
        """Возвращает адресные объекты, соответствующие параметрам поиска.

        :param unicode filter_string: Строка поиска.
        :param levels: Уровни адресных объектов, среди которых нужно осуществлять поиск.
        :param typenames: Наименования типов адресных объектов, среди которых нужно осуществлять поиск.
        :param parent_id: ID родительского объекта.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: generator
        """
        return find_address_objects(filter_string, levels, typenames, parent_id, timeout)

    def get_address_object(self, obj_id, timeout=None):
        """Возвращает адресный объект ГАР по его ID.

        :param obj_id: ID адресного объекта ГАР.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.AddressObject
        """
        return get_address_object(obj_id, timeout)

    def find_house(self, house_number, parent_id, timeout=None):
        """Возвращает информацию о здании по его номеру.

        :param unicode house_number: Номер дома.
        :param parent_id: ID родительского объекта.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.House or NoneType
        """
        return find_house(house_number, parent_id, timeout)

    def get_house(self, house_id, timeout=None):  # pylint: disable=signature-differs
        """Возвращает информацию о здании по его ID в ГАР.

        :param house_id: ID здания.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.House
        """
        return get_house(house_id, timeout)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from m3_gar_client.constants import GAR_HIERARCHIES
from m3_gar_client.constants import GAR_HIERARCHY_MUN
from m3_gar_client.data import AddressObject
from m3_gar_client.data import House


class cached_property(property):
    """Кешируемое свойство.

    В отличие от :class:`django.utils.functional.cached_property`, наследуется
    от property и копирует строку документации, что актуально при генерации
    документации средствами Sphinx.
    """

    def __init__(self, method):
        super(cached_property, self).__init__(method)

        self.__doc__ = method.__doc__

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.fget.__name__ not in instance.__dict__:
            instance.__dict__[self.fget.__name__] = self.fget(instance)

        return instance.__dict__[self.fget.__name__]


def correct_keyboard_layout(text):
    """При необходимости меняет раскладку клавиатуры.

    :param unicode text: Текстовая строка, подлежащая корректировке.

    :rtype: unicode
    """
    en_chars = (
        '`~@#$^&'
        'qwertyuiop[]'
        'QWERTYUIOP{}'
        'asdfghjkl;\''
        'ASDFGHJKL:"|'
        'zxcvbnm,./'
        'ZXCVBNM<>?'
    )

    ru_chars = (
        'ёЁ"№;:?'
        'йцукенгшщзхъ'
        'ЙЦУКЕНГШЩЗХЪ'
        'фывапролджэ'
        'ФЫВАПРОЛДЖЭ/'
        'ячсмитьбю.'
        'ЯЧСМИТЬБЮ,'
    )

    assert len(en_chars) == len(ru_chars)

    ru_only_chars = set(ru_chars) - set(en_chars)
    if set(text).isdisjoint(ru_only_chars):
        # Текст не содержит ни одного символа из русской раскладки, значит
        # раскладку надо поменять.

        def translate():
            for char in text:
                position = en_chars.find(char)
                yield char if position == -1 else ru_chars[position]

        result = ''.join(translate())
    else:
        result = text

    return result


def find_address_objects(filter_string, levels=None, typenames=None, parent_id=None, timeout=None):
    """Возвращает адресные объекты, соответствующие параметрам поиска.

    :param unicode filter_string: Строка поиска.
    :param levels: Уровни адресных объектов, среди которых нужно осуществлять поиск.
    :param parent_id: ID родительского объекта.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: generator
    """
    from m3_gar_client import config

    return config.backend.find_address_objects(filter_string, levels, typenames, parent_id, timeout)


def get_address_object(obj_id, timeout=None):
    """Возвращает адресный объект ГАР по его ID.

    :param obj_id: ID адресного объекта ГАР.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: m3_gar_client.data.AddressObject
    """
    from m3_gar_client import config

    return config.backend.get_address_object(obj_id, timeout)


def find_house(house_number='', parent_id=None, timeout=None):
    """Возвращает информацию о здании по его номеру.

    :param unicode house_number: Номер дома.
    :param parent_id: ID родительского объекта.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: m3_gar_client.data.House or NoneType
    """
    from m3_gar_client import config

    return config.backend.find_house(house_number, parent_id, timeout)


def get_house(house_id, parent_id=None, timeout=None):
    """Возвращает информацию о здании по его ID в ГАР.

    .. important::

       В ГАР здания с разными корпусами/строениями имеют разные ID.
       Например, "д.1 корп. 1" и "д.1 корп. 2" будут иметь разные
       идентификаторы.

    :param house_id: ID здания.
    :param parent_id: ID адресного объекта, в котором находится здание.
        Необходимость указывать ID определяется используемым бэкендом.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: m3_gar_client.data.House
    """
    from m3_gar_client import config

    return config.backend.get_house(house_id, parent_id, timeout)


def get_address_object_name(address_object):
    """Возвращает наименование объекта с кратким наименованием его типа.

    Примеры:

      * Забайкальский край
      * ул. Ленина
      * г. Казань

    :type address_object: m3_gar_client.data.AddressObject

    :rtype: unicode
    """
    if address_object.guid == 'd66e5325-3a25-4d29-ba86-4ca351d9704b':
        # Ханты-Мансийский Автономный округ - Югра
        result = address_object.formal_name
    elif address_object.short_name in ('край', 'АО', 'Аобл', 'обл'):
        result = '{} {}'.format(address_object.formal_name, address_object.short_name)
    else:
        result = '{}. {}'.format(address_object.short_name, address_object.formal_name)

    return result


def get_house_name(house):
    """Возвращает полное наименование здания.

    Примеры:

      * д. 1
      * д. 2 корп. 3
      * корп. 5
      * д. 4 стр. 5
      * стр. 9
      * д. 6 корп. 7 стр. 8

    :type house: m3_gar_client.data.House

    :rtype: unicode
    """
    assert isinstance(house, House), type(house)

    names = []
    if house.house_number:
        names.append('д. ' + house.house_number)
    if house.building_number:
        names.append('корп. ' + house.building_number)
    if house.structure_number:
        names.append('стр. ' + house.structure_number)

    return ', '.join(names)


def get_full_name(obj, hierarchy=GAR_HIERARCHY_MUN, with_postal_code=True, timeout=None):
    """Возвращает полное наименование адресного объекта или здания.

    Примеры:

      * Забайкальский край, г. Чита
      * Новосибирская обл., г. Новосибирск, ул. Вокзальная магистраль, д. 1/1
      * д. 1 корп. 3 стр. 2

    :type obj: m3_gar_client.data.AddressObject or m3_gar_client.data.House
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: unicode
    """
    postal_code = None
    names = []

    if not isinstance(obj, (House, AddressObject)):
        raise TypeError(obj)

    for addrobj in get_full_details(obj, hierarchy=hierarchy):
        if postal_code is None and addrobj.postal_code:
            postal_code = addrobj.postal_code

        if isinstance(addrobj, House):
            names.append(get_house_name(addrobj))

        elif isinstance(addrobj, AddressObject):
            names.append(get_address_object_name(addrobj))

    if with_postal_code and postal_code is not None:
        names.insert(0, postal_code)

    full_name = ', '.join(reversed(names))

    return full_name


def get_full_details(obj, hierarchy=GAR_HIERARCHY_MUN, timeout=None):
    """
    Возвращает полный список объектов, снизу вверх из иерархии.
    """
    if not isinstance(obj, (House, AddressObject)):
        raise TypeError(obj)

    if hierarchy not in GAR_HIERARCHIES:
        raise ValueError(hierarchy)

    parent_attr = f'{hierarchy}_parent_obj_id'

    objects = []

    if isinstance(obj, House):
        objects.append(obj)

        parent_id = getattr(obj, parent_attr)
        if parent_id:
            obj = get_address_object(parent_id, timeout)

    if isinstance(obj, AddressObject):
        while obj:
            objects.append(obj)

            parent_id = getattr(obj, parent_attr)
            if parent_id:
                obj = get_address_object(parent_id, timeout)
            else:
                break

    return objects


def get_details(obj, timeout=None):
    """
    Возвращает список наименований и сокращений узла, снизу вверх из иерархии.
    Не включая верхний уровень "Субъект РФ"
    """
    result = get_full_details(obj, timeout=timeout)

    return result[1:] if len(result) > 1 else result



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from abc import ABCMeta
from abc import abstractmethod
from abc import abstractproperty

from six import with_metaclass


class BackendBase(with_metaclass(ABCMeta, object)):

    """Базовый класс для бэкендов."""

    @abstractproperty
    def place_search_url(self):
        """URL для поиска населенных пунктов.

        :rtype: str
        """

    @abstractproperty
    def street_search_url(self):
        """URL для поиска улиц.

        :rtype: str
        """

    @abstractproperty
    def house_search_url(self):
        """URL для запроса списка домов.

        :rtype: str
        """

    def configure_place_field(self, field):
        """Настраивает поле "Населенный пункт".

        :param field: Поле "Населенный пункт".
        :type field: m3_ext.ui.fields.simple.ExtComboBox
        """

    def configure_street_field(self, field):
        """Настраивает поле "Улица".

        :param field: Поле "Улица".
        :type field: m3_ext.ui.fields.simple.ExtComboBox
        """

    def configure_house_field(self, field):
        """Настраивает поле "Дом".

        :param field: Поле "Дом".
        :type field: m3_ext.ui.fields.simple.ExtComboBox
        """

    @abstractmethod
    def find_address_objects(
        self,
        filter_string,
        levels=None,
        typenames=None,
        parent_id=None,
        timeout=None,
    ):
        """Возвращает адресные объекты, соответствующие параметрам поиска.

        :param unicode filter_string: Строка поиска.
        :param levels: Уровни адресных объектов, среди которых нужно осуществлять поиск.
        :param parent_id: ID родительского объекта.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: generator
        """

    @abstractmethod
    def get_address_object(self, obj_id, timeout=None):
        """Возвращает адресный объект ГАР по его ID.

        :param obj_id: ID адресного объекта ГАР.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.AddressObject
        """

    @abstractmethod
    def find_house(self, house_number, parent_id, timeout=None):
        """Возвращает информацию о здании по его номеру.

        :param unicode house_number: Номер дома.
        :param parent_id: ID родительского объекта.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.House or NoneType
        """

    @abstractmethod
    def get_house(self, house_id, timeout=None):
        """Возвращает информацию о здании по его ID в ГАР.

        :param house_id: ID здания.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.House
        """



# coding: utf-8
from os.path import abspath
from os.path import dirname
from os.path import join
import re

from pkg_resources import Requirement
from setuptools import find_packages
from setuptools import setup


_COMMENT_RE = re.compile(r'(^|\s)+#.*$')


def _get_requirements(file_path):
    with open(file_path, 'r') as file:
        for line in file:
            line = _COMMENT_RE.sub('', line)
            line = line.strip()

            if line.startswith('-r '):
                for req in _get_requirements(join(dirname(abspath(file_path)), line[3:])):
                    yield req

            elif line:
                req = Requirement(line)
                req_str = req.name + str(req.specifier)

                if req.marker:
                    req_str += '; ' + str(req.marker)

                yield req_str


def _read(file_path):
    with open(file_path, 'r') as infile:
        return infile.read()


setup(
    name='m3-gar-client',
    url='https://stash.bars-open.ru/projects/M3/repos/m3-gar-client',
    license='MIT',
    author='BARS Group',
    description=u'UI клиент для сервера ГАР m3-rest-gar',
    author_email='bars@bars-open.ru',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=tuple(_get_requirements('requirements/prod.txt')),
    long_description=_read('README.rst'),
    include_package_data=True,
    classifiers=[
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Natural Language :: Russian',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django :: 1.4',
        'Framework :: Django :: 1.5',
        'Framework :: Django :: 1.6',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
    ],
    extras_require={
        'oauth2': (
            'oauthlib>=2,<3',
            'requests-oauthlib<1',
        ),
    },
    dependency_links=(
        'http://pypi.bars-open.ru/simple/m3-builder',
    ),
    setup_requires=(
        'm3-builder>=1.2,<2',
    ),
    set_build_info=dirname(__file__),
)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from abc import ABCMeta
from abc import abstractmethod

from django.http.response import JsonResponse
from m3.actions import Action
from m3.actions import ActionPack
from six import text_type
from six import with_metaclass

from m3_gar_client.backends.m3_rest_gar.proxy.utils import PlaceLoader
from m3_gar_client.backends.m3_rest_gar.proxy.utils import StreetLoader
from m3_gar_client.backends.m3_rest_gar.proxy.utils import UIHouseLoader
from m3_gar_client.utils import correct_keyboard_layout


class ActionBase(with_metaclass(ABCMeta, Action)):
    """
    Базовый класс для обработчиков запросов на поиск данных в ГАР.
    """

    @abstractmethod
    def _get_loader(self, context):
        """Возвращает загрузчик данных.

        :rtype: m3_gar_client.backends.m3_rest_gar.proxy.utils.LoaderBase
        """

    def context_declaration(self):
        return {
            'filter': {
                'type': 'unicode',
                'default': '',
            },
        }

    def run(self, request, context):
        context.filter = correct_keyboard_layout(context.filter)

        loader = self._get_loader(context)
        rows = loader.load()

        return JsonResponse({
            'rows': rows,
            'total': len(rows),
        })


class PlaceSearchAction(ActionBase):
    """
    Обработчик запросов на поиск населенных пунктов.
    """

    url = '/search/place'

    def _get_loader(self, context):
        # pylint: disable=abstract-class-instantiated
        return PlaceLoader(context.filter)


class ParentMixin(object):

    def context_declaration(self):
        result = super(ParentMixin, self).context_declaration()

        result.update({
            'parent': {
                'type': 'unicode',
            },
        })

        return result


class StreetSearchAction(ParentMixin, ActionBase):
    """
    Обработчик запросов на поиск улиц в населенном пункте.
    """

    url = '/search/street'

    def _get_loader(self, context):
        # pylint: disable=abstract-class-instantiated
        return StreetLoader(context.filter, parent_id=text_type(context.parent))


class HouseSearchAction(ParentMixin, ActionBase):
    """
    Обработчик запросов на поиск домов.
    """

    url = '/search/house'

    def _get_loader(self, context):
        # pylint: disable=abstract-class-instantiated
        return UIHouseLoader(context.filter, parent_id=context.parent)


class Pack(ActionPack):
    """
    Набор действий для проксирования запросов к серверу ГАР.
    """

    url = '/gar'

    def __init__(self):
        super(Pack, self).__init__()

        self.place_search_action = PlaceSearchAction()
        self.street_search_action = StreetSearchAction()
        self.house_search_action = HouseSearchAction()

        self.actions.extend((
            self.place_search_action,
            self.street_search_action,
            self.house_search_action,
        ))



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from abc import ABCMeta
from abc import abstractmethod
from itertools import chain
from itertools import count

from six import iteritems
from six import text_type
from six import with_metaclass
from six.moves import http_client
from six.moves import map
from six.moves import range

from m3_gar_client.backends.m3_rest_gar.proxy.server import server
from m3_gar_client.constants import GAR_HIERARCHY_MUN
from m3_gar_client.constants import GAR_LEVELS_PLACE
from m3_gar_client.constants import GAR_LEVELS_STREET
from m3_gar_client.data import AddressObject
from m3_gar_client.data import House
from m3_gar_client.data import ObjectMapper
from m3_gar_client.utils import cached_property


# -----------------------------------------------------------------------------
# Обёртки над данными, поступающими с сервера m3-rest-gar.

class AddressObjectMapper(ObjectMapper):
    """
    Обертка над данными сервера m3-rest-gar об адресных объектах.

    Предназначена для использования при создании экземпляров
    :class:`m3_gar_client.data.AddressObject`.
    """

    fields_map = {
        'id': 'id',
        'guid': 'objectguid',
        'obj_id': 'objectid',
        'previous_id': 'previd',
        'next_id': 'nextid',
        'ifns_fl_code': 'params__IFNSFL__value',
        'ifns_fl_territorial_district_code': 'params__territorialifnsflcode__value',
        'ifns_ul_code': 'params__IFNSUL__value',
        'ifns_ul_territorial_district_code': 'params__territorialifnsulcode__value',
        'okato': 'params__OKATO__value',
        'oktmo': 'params__OKTMO__value',
        'postal_code': 'params__PostIndex__value',
        'formal_name': 'name',
        'official_name': 'name',
        'short_name': 'typename',
        'level': 'level',
        'kladr_code': 'params__CODE__value',
        'kladr_plain_code': 'params__PLAINCODE__value',
        'date_of_update': 'updatedate',
        'date_of_creation': 'startdate',
        'date_of_expiration': 'enddate',
        'full_name': lambda data: '{} {}'.format(data['name'], data['typename']),
        'region_code': 'region_code',
        'adm_parent_obj_id': 'hierarchy__adm__parentobjid__objectid',
        'adm_parent_guid': 'hierarchy__adm__parentobjid__objectguid',
        'mun_parent_obj_id': 'hierarchy__mun__parentobjid__objectid',
        'mun_parent_guid': 'hierarchy__mun__parentobjid__objectguid',
    }

    assert set(fields_map) == set(AddressObject.fields)


class HouseMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar об адресных объектах.

    Предназначена для использования при создании экземпляров
    :class:`m3_gar_client.data.House`.
    """

    fields_map = {
        'id': 'id',
        'guid': 'objectguid',
        'obj_id': 'objectid',
        'ifns_fl_code': 'params__IFNSFL__value',
        'ifns_fl_territorial_district_code': 'params__territorialifnsflcode__value',
        'ifns_ul_code': 'params__IFNSUL__value',
        'ifns_ul_territorial_district_code': 'params__territorialifnsulcode__value',
        'okato': 'params__OKATO__value',
        'oktmo': 'params__OKTMO__value',
        'postal_code': 'params__PostIndex__value',
        'house_number': 'housenum',
        'building_number': 'addnum1',
        'structure_number': 'addnum2',
        'date_of_update': 'updatedate',
        'date_of_creation': 'startdate',
        'date_of_end': 'enddate',
        'region_code': 'region_code',
        'adm_parent_obj_id': 'hierarchy__adm__parentobjid__objectid',
        'adm_parent_guid': 'hierarchy__adm__parentobjid__objectguid',
        'mun_parent_obj_id': 'hierarchy__mun__parentobjid__objectid',
        'mun_parent_guid': 'hierarchy__mun__parentobjid__objectguid',
    }

    assert set(fields_map) == set(House.fields)


class UIAddressObjectMapper(ObjectMapper):
    """
    Обертка над данными сервера m3-rest-gar об адресных объектах.
    Предназначена для использования в UI.
    """

    fields_map = {
        'objectId': 'objectguid',
        'level': 'level',
        'shortName': 'typename',
        'formalName': 'name',
        'postalCode': 'params__PostIndex__value',
        'fullName': lambda data: '{} {}'.format(data['name'], data['typename']),
    }


class UIHouseMapper(ObjectMapper):
    """
    Обертка над данными сервера m3-rest-gar о зданиях.
    Предназначена для использования в UI.
    """

    fields_map = {
        'objectId': 'objectguid',
        'houseNumber': 'housenum',
        'buildingNumber': 'addnum1',
        'structureNumber': 'addnum2',
        'postalCode': 'params__PostIndex__value',
    }

# -----------------------------------------------------------------------------
# Загрузчики данных с сервера m3-rest-gar.


class LoaderBase(with_metaclass(ABCMeta, object)):
    """
    Базовый класс для загрузчиков объектов ГАР.
    """

    def __init__(self, filter_string, **kwargs):
        """Инициализация экземпляра класса.

        :param unicode filter_string: Строка для фильтрации объектов.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.
        """
        self.filter_string = filter_string
        self.timeout = kwargs.get('timeout')
        self.exact = kwargs.get('exact', False)

    @property
    @abstractmethod
    def _path(self):
        """Путь к ресурсу API сервера ГАР.

        :rtype: str
        """

    @cached_property
    def _fields(self):
        """Имена полей, подлежащих загрузке.

        :rtype: tuple
        """
        return list(self._mapper_class.fields_map.keys())

    @property
    @abstractmethod
    def _mapper_class(self):
        """Класс, преобразующий имена полей.

        ::rtype: m3_gar_client.data.ObjectMapper
        """

    @property
    def _filter_param(self):
        """Query-параметр для фильтрации.

        :rtype: str
        """
        return 'name'

    def _load_page(self, params, page):
        params = params.copy()
        params['page'] = page

        drf_response = server.get(self._path, params, timeout=self.timeout)
        if drf_response.status_code == http_client.OK:
            result = drf_response.json()
        else:
            result = None

        return result

    def _process_object_data(self, drf_object_data):
        """Выполняет дополнительную обработку данных объекта ГАР.

        :param dict drf_object_data: Данные объекта ГАР, полученные с сервера m3-rest-gar.
        """
        return map_object_data(self._mapper_class, drf_object_data)

    def _build_result(self, object_data):
        """Формирует данные результирующего объекта ГАР.

        Полученные данные включаются в результат загрузки.

        :param dict object_data: Данные объекта, полученные с сервера ГАР и
            прошедшие обработку в методе ``_process_object_data``.

        :rtype: dict
        """
        return {
            field: object_data[field]
            for field in self._fields
        }

    @abstractmethod
    def _filter(self, object_data):
        """Возвращает True, если объект ГАР должен попасть в загрузку.

        :rtype: bool
        """

    def _get_params(self):
        """Возвращает параметры запроса к серверу ГАР.

        :rtype: dict
        """
        filter_param = self._filter_param

        if self.exact:
            filter_param += '__exact'

        result = {
            filter_param: self.filter_string,
        }

        return result

    def load_raw(self, page=None):
        """Возвращает данные адресных объектов в исходном виде.

        :param int page: Номера страницы для загрузки данных. Значение ``None``
            указывает на необходимость загрузки всех страниц.

        :rtype: generator
        """
        if page is None:
            pages = count(start=1)
        else:
            pages = range(page, page + 1)

        params = self._get_params()

        for page_number in pages:
            drf_data = self._load_page(params, page_number)
            if drf_data:
                for drf_object_data in drf_data['results']:
                    object_data = self._process_object_data(drf_object_data)
                    if self._filter(object_data):
                        yield object_data

                if not drf_data['next']:
                    break
            else:
                break

    def load_results(self, page=None):
        """Возвращает данные в соответствии с параметрами загрузчика.

        :param int page: Номера страницы для загрузки данных. Значение ``None``
            указывает на необходимость загрузки всех страниц.

        :rtype: itertools.imap
        """
        return map(self._build_result, self.load_raw(page))

    @abstractmethod
    def _sort_key(self, object_data):
        """Возвращает значение ключа для сортировки результатов загрузки.

        :param dict object_data: Данные загруженного объекта ГАР.
        """

    def _process_result(self, data):
        """Обработка полученных после загрузки данных.

        :param tuple data: Кортеж словарей с данными загруженных объектов ГАР.
        """
        return sorted(data, key=self._sort_key)

    def load(self, page=None):
        """Загружает данные с сервера ГАР.

        :param int page: Номера страницы для загрузки данных. Значение ``None``
            указывает на необходимость загрузки всех страниц.

        :rtype: collections.Iterable
        """
        data = tuple(self.load_results(page))
        result = self._process_result(data)

        return result


class ParentFilterMixin(object):
    """
    Миксин для загрузчиков, которые используют id родительского объекта в запросах к серверу ГАР
    """

    @property
    @abstractmethod
    def _default_hierarchy(self):
        """Вид иерархии, используемый по умолчанию"""

    def __init__(self, *args, **kwargs):
        super(ParentFilterMixin, self).__init__(*args, **kwargs)

        self._hierarchy = kwargs.get('hierarchy') or self._default_hierarchy
        self._parent_id = kwargs.get('parent_id')

    def _get_params(self):
        result = super(ParentFilterMixin, self)._get_params()

        if self._parent_id:
            result['parent'] = text_type('{}:{}'.format(self._hierarchy, self._parent_id))

        return result


class AddressObjectLoaderBase(ParentFilterMixin, LoaderBase):
    """Базовый класс для загрузчиков адресных объектов ГАР.

    В терминологии ГАР адресными объектами называются:

        * субъекты РФ;
        * административные и муниципальные районы субъектов РФ;
        * города;
        * сельские/городские поселения;
        * населенные пункты;
        * элементы планировочной структуры;
        * улицы.
    """

    _default_hierarchy = GAR_HIERARCHY_MUN

    @property
    def _path(self):
        return '/addrobj/'

    @property
    @abstractmethod
    def _levels(self):
        """Уровни адресных объектов, для которых нужно искать данные в ГАР.

        :rtype: tuple
        """

    def _get_params(self):
        """Возвращает параметры запроса к серверу ГАР.

        :rtype: dict
        """
        result = super(AddressObjectLoaderBase, self)._get_params()

        if self._levels:
            result['level'] = ','.join(map(str, self._levels))

        if self._typenames:
            result['typename'] = ','.join(map(str, self._typenames))

        return result

    def _filter(self, object_data):
        return True

    def _sort_key(self, object_data):
        return object_data['fullName'] or ''


class AddressObjectLoader(AddressObjectLoaderBase):
    """Загрузчик адресных объектов ГАР.

    Загружает информацию об адресных объектах, соответствующих строке
    фильтрации и находящихся на одном из указанных в параметрах экземпляра
    уровней иерархии адресных объектов.
    """

    _levels = None

    _mapper_class = AddressObjectMapper

    def __init__(self, filter_string, levels=None, typenames=None, **kwargs):
        """Инициализация экземпляра класса.

        :param unicode filter_string: Строка для фильтрации объектов.
        :param levels: Уровни адресных объектов.
        """
        super(AddressObjectLoader, self).__init__(filter_string, **kwargs)

        self._levels = levels
        self._typenames = typenames


class PlaceLoader(AddressObjectLoaderBase):
    """
    Загрузчик сведений о населенных пунктах
    """

    _levels = GAR_LEVELS_PLACE

    _mapper_class = UIAddressObjectMapper


class StreetLoader(AddressObjectLoaderBase):
    """
    Загрузчик сведений об улицах.
    """

    _levels = GAR_LEVELS_STREET

    _mapper_class = UIAddressObjectMapper


class HouseLoader(ParentFilterMixin, LoaderBase):
    """
    Загрузчик сведений о зданиях.
    """

    _default_hierarchy = GAR_HIERARCHY_MUN

    _mapper_class = HouseMapper

    @property
    def _path(self):
        return '/houses/'

    @property
    def _filter_param(self):
        return 'housenum'

    def _filter(self, object_data):
        return True

    def _sort_key(self, object_data):
        return object_data['house_number'] or ''

    @staticmethod
    def _split_number(number):
        """Разделяет номер на целочисленную и буквенную части.

        :param unicode number: Номер дома/корпуса/строения.

        :rtype: tuple
        """
        int_part = ''.join(ch for ch in number if ch.isdigit())
        str_part = number[len(int_part):]

        return int(int_part) if int_part else -1, str_part


class UIHouseLoader(HouseLoader):
    _mapper_class = UIHouseMapper

    def _sort_key(self, object_data):
        return tuple(chain(*(
            self._split_number(object_data[number_type + 'Number'])
            for number_type in ('house', 'building', 'structure')
        )))

    def _filter(self, object_data):
        """Возвращает True для записей, соответствующих параметрам поиска.

        Запись считается соответствующей указанным при инициализации загрузчика
        параметрам поиска, если:

            * номер дома (если есть) в записи **начинается со строки**,
              указанной в аргументе ``filter_string``;
            * номер корпуса или строения (если номер дома отсутствует) в записи
              **начинается со строки**, указанной в аргументе
              ``filter_string``;
            * в аргументе ``filter_string`` конструктора класса было передано
              значение ``None``.

        :rtype: bool
        """
        if self.filter_string is None:
            result = True
        else:
            house = (object_data.get('houseNumber') or '').lower()

            if house:
                result = house.startswith(self.filter_string)
            else:
                building = (object_data.get('buildingNumber') or '').lower()
                structure = (object_data.get('structureNumber') or '').lower()
                result = (
                    building and building.startswith(self.filter_string) or
                    structure and structure.startswith(self.filter_string)
                )
        return result

    def _process_object_data(self, drf_object_data):
        house_data = super(UIHouseLoader, self)._process_object_data(drf_object_data)

        house_data['houseNumber'] = house_data['houseNumber'] or ''
        house_data['buildingNumber'] = house_data['buildingNumber'] or ''
        house_data['structureNumber'] = house_data['structureNumber'] or ''

        return house_data

# -----------------------------------------------------------------------------
# Функции для создания объектов m3-gar-client на основе данных m3-rest-gar.

def get_address_object(obj_id, timeout=None):
    """Возвращает адресный объект, загруженный с сервера ГАР.

    Если адресный объект не найден, возвращает ``None``.

    :param obj_id: ObjectID/GUID адресного объекта ГАР.
    :param float timeout: timeout запроса к серверу ГАР в секундах.

    :rtype: m3_gar_client.data.AddressObject or NoneType

    :raises requests.ConnectionError: если не удалось соединиться с сервером ГАР
    """
    assert obj_id is not None

    response = server.get('/addrobj/{}/'.format(obj_id), timeout=timeout)

    if response.status_code == http_client.OK:
        response_data = response.json()
        mapped_data = map_object_data(AddressObjectMapper, response_data)
        result = AddressObject(**mapped_data)
    else:
        result = None

    return result


def find_address_objects(
    filter_string,
    levels=None,
    typenames=None,
    parent_id=None,
    timeout=None,
):
    """Возвращает адресные объекты, соответствующие параметрам поиска.

    :param unicode filter_string: Строка поиска.
    :param levels: Уровни адресных объектов, среди которых нужно осуществлять поиск.
    :param parent_id: ID родительского объекта.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: generator
    """
    return AddressObjectLoader(
        filter_string,
        levels=levels,
        typenames=typenames,
        parent_id=parent_id,
        timeout=timeout,
    ).load_results()


def get_house(house_id, timeout=None):
    """Возвращает информацию о здании по его ID в ГАР.

    :param house_id: ID здания.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: m3_gar_client.data.House
    """
    assert house_id is not None

    response = server.get('/houses/{}/'.format(house_id), timeout=timeout)

    if response.status_code == http_client.OK:
        response_data = response.json()
        mapped_data = map_object_data(HouseMapper, response_data)
        result = House(**mapped_data)
    else:
        result = None

    return result


def find_house(house_number='', parent_id=None, timeout=None):
    """Возвращает информацию о здании по его номеру.

    :param unicode house_number: Номер дома.
    :param parent_id: ID родительского объекта.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: m3_gar_client.data.House or NoneType
    """
    houses = HouseLoader(
        house_number,
        parent_id=parent_id,
        exact=True,
        timeout=timeout,
    ).load()

    if len(houses) == 1:
        result = House(**houses[0])
    else:
        result = None

    return result


def flatten_object_data(raw_object_data):
    """Преобразует данные объекта ГАР к "плоскому" виду

    :param dict raw_object_data: Необработанные данные объекта ГАР, полученные с сервера m3-rest-gar.
    """
    flat_data = {}

    for key, value in iteritems(raw_object_data):
        if type(value) is dict:
            for inner_key, inner_value in iteritems(flatten_object_data(value)):
                flat_data['{}__{}'.format(key, inner_key)] = inner_value

        elif type(value) is list and key == 'params':
            for param_dict in value:
                if type(param_dict) is not dict:
                    continue

                code = param_dict.get('typeid', {}).get('code')
                if code is None:
                    continue

                for inner_key, inner_value in iteritems(flatten_object_data(param_dict)):
                    flat_data['params__{}__{}'.format(code, inner_key)] = inner_value

        else:
            flat_data[key] = value

    return flat_data


def map_object_data(mapper_class, raw_object_data):
    """Выполняет маппинг сырых данных, полученных с сервера m3-rest-gar

    :param dict raw_object_data: Необработанные данные объекта ГАР, полученные с сервера m3-rest-gar.
    :param class ObjectMapper mapper_class: класс-маппер.
    """
    flat_data = flatten_object_data(raw_object_data)
    mapped_data = mapper_class(flat_data)

    return mapped_data
